from numbers import Number as numeric
from charcad.draw.coordinates import Coordinates
from charcad.draw.graph import GraphicObject, Graph
from charcad.draw.utils import Formatter


class Point(GraphicObject):
    def __init__(self, x=0, y=0, marker='.', formatter=None):
        if isinstance(y, str):
            marker = y
            y = None
        super().__init__(x, y)
        if not isinstance(marker, str):
            raise TypeError("marker should be str.")
        self.marker = marker[0]
        self.set_formatter(formatter)

    def __add__(self, other):
        if isinstance(other, Point):
            new_point = Point(self.coord + other.coord, marker=self.marker,
                              formatter=self.formatter)
        elif isinstance(other, (Coordinates, tuple)):
            new_point = Point(self.coord + other, marker=self.marker,
                              formatter=self.formatter)
        else:
            raise TypeError("'other' must be a Point or a Coordinates object.")
        return new_point

    def __sub__(self, other):
        if isinstance(other, Point):
            new_point = Point(self.coord - other.coord, marker=self.marker,
                              formatter=self.formatter)
        elif isinstance(other, (Coordinates, tuple)):
            new_point = Point(self.coord - other, marker=self.marker,
                              formatter=self.formatter)
        else:
            raise TypeError("'other' must be a Point or a Coordinates object.")
        return new_point

    def __mul__(self, other):
        if isinstance(other, numeric):
            return Point(self.coord*other, marker=self.marker,
                         formatter=self.formatter)
        else:
            raise TypeError("'other' must be numeric.")

    def __truediv__(self, other):
        if isinstance(other, numeric):
            return Point(self.coord/other, marker=self.marker,
                         formatter=self.formatter)
        else:
            raise TypeError("'other' must be numeric.")

    def __eq__(self, other):
        if isinstance(other, Point):
            eq = self.coord == other.coord
        elif isinstance(other, Coordinates):
            eq = self.coord == other
        else:
            eq = False
        return eq

    def __getitem__(self, item):
        return self.coord[item]

    def __repr__(self):
        return 'Point({}, {}, {})'.format(self.x, self.y, self.marker)

    def __floor__(self):
        return Point(int(self.coord), marker=self.marker,
                     formatter=self.formatter)

    def __round__(self, n=0):
        return Point(round(self.coord, n), marker=self.marker,
                     formatter=self.formatter)

    def set_formatter(self, formatter):
        if formatter is None:
            formatter = Formatter()
        self.formatter = formatter
        self.set_marker(self.marker)

    def set_marker(self, marker):
        self.marker = marker[0]
        self.graph = Graph(self.formatter.format(self.marker))



from charcad.version import __version__

from charcad.draw.coordinates import Coordinates as Coord
from charcad.draw.canvas import Canvas
from charcad.draw.block import Block
from charcad.draw.point import Point
from charcad.draw.lines import Horizontal, Vertical, Route
from charcad.draw.text import Text
from charcad.draw.graph import Graph
from charcad.draw.utils import Formatter



import math
import numpy as np
from copy import deepcopy

from charcad.draw.coordinates import Coordinates
from charcad.draw.utils import DrawingChars


_EMPTY_GRID = np.array([[' ']])


class GraphicObject:
    def __init__(self, x=0, y=0, transparent=True):
        self.set_coordinates(x, y)
        self.graph = None
        self.set_transparency(transparent)

    @property
    def x(self):
        return self.coord.x

    @property
    def y(self):
        return self.coord.y

    def set_coordinates(self, x, y=None):
        if isinstance(x, Coordinates):
            coord = x
        elif isinstance(x, (list, tuple)):
            coord = Coordinates(*x)
        else:
            coord = Coordinates(x, y)
        self.coord = coord

    def set_transparency(self, transparent):
        self.transparent = transparent

    def set_x(self, x):
        self.coord.x = x

    def set_y(self, y):
        self.coord.y = y

    def show(self, axes=False, frame=False, frame_formatter=None):
        self.graph.print(axes=axes, frame=frame,
                         frame_formatter=frame_formatter)


class GraphicObjectArray:
    def __init__(self):
        self._objects = dict()
        self._counter = 0

    def __getitem__(self, key):
        if isinstance(key, str):
            self._objects[key]
        else:
            key_aux = self.keys
            key_aux.sort()
            if isinstance(key, slice):
                item = [self._objects[k] for k in key_aux[key]]
            else:
                item = self._objects[key_aux[key]]
        return item

    def __len__(self):
        return len(self._objects)

    def __repr__(self):
        z = zip(self.keys, self.values)
        p = int(math.log10(len(self)) + 1)
        return 'Graphic Object Array[\n  %s\n]' % ',\n  '\
            .join([(str(i).zfill(p) + '\t' + k + ': %s') % v
                   for i, (k, v) in enumerate(z)])

    @property
    def keys(self):
        return list(self._objects.keys())

    @property
    def values(self):
        return list(self._objects.values())

    def add(self, obj):
        name = 'object_{:04d}'.format(self._counter)
        self._check_duplicates(obj)
        self._objects.update({name: obj})
        self._counter += 1

    def _check_duplicates(self, obj):
        [delete_object(o) for o in self if obj == o]

    def remove(self, key):
        if isinstance(key, (list, tuple)):
            key = list(key)
            key.sort(reverse=True)
            [self.remove(k) for k in key]
        else:
            if isinstance(key, str):
                pass
            elif isinstance(key, int):
                key = self.keys[key]
            else:
                raise TypeError("key must be integer, string or iterable.")
            del self._objects[key]

    def update(self, item):
        self._objects.update(item)


class Graph:
    def __init__(self, grid=None, w=0, h=0):
        if grid is None:
            grid = np.full((h, w), ' ', dtype=object)
        elif isinstance(grid, str):
            grid = np.array([[grid]])
        elif isinstance(grid, np.ndarray):
            pass
        else:
            raise TypeError('grid cant be numeric.')
        self.grid = grid

    def __repr__(self):
        return '\n'.join([''.join(row) for row in self.grid])

    def __len__(self):
        return len(self.grid)

    def __getitem__(self, key):
        return Graph(self.grid[key])

    def __setitem__(self, key, value):
        # i utterly hate this
        key = list(key)
        for i, k in enumerate(key):
            if not isinstance(k, slice):
                key[i] = slice(k, k+1)
        key = tuple(key)
        self.grid[key] = value.grid

    @property
    def h(self):
        return self.grid.shape[0]

    @property
    def w(self):
        return self.grid.shape[1]

    @property
    def shape(self):
        return (self.h, self.w)

    def add_objects(self, graphicArray):
        for obj in graphicArray:
            if isinstance(obj, GraphicObjectArray):
                self.add_objects(obj)
            elif isinstance(obj, GraphicObject):
                self.add_graph(obj)
            else:
                raise TypeError(
                    "obj must be GraphicObject or GraphicObjectArray.")

    def add_graph(self, obj, x=0, y=0):
        if isinstance(x, Coordinates):
            coord_other = x
        else:
            coord_other = Coordinates(x, y)
        coord = obj.coord + coord_other
        xi = coord.x
        xf = coord.x + obj.graph.w
        yf = self.flipud(coord.y - 1)
        yi = self.flipud(coord.y - 1 + obj.graph.h)
        x_slice = slice(xi, xf)
        y_slice = slice(yi, yf)
        transparent = obj.transparent
        if transparent:
            new_subgraph = self[y_slice, x_slice].no_background_draw(obj.graph)
        else:
            new_subgraph = obj.graph
        self[y_slice, x_slice] = new_subgraph

    def copy(self):
        grid = self.grid.copy()
        return Graph(grid=grid)

    def flipud(self, yi):
        return self.h - yi - 1

    def inspect(self):
        [print(item) for item in [row for row in self.grid]]

    def isempty(self):
        return self.grid == _EMPTY_GRID

    def no_background_draw(self, other, rng=None):
        gph_out = self.copy()
        for i in range(self.h):
            for j in range(self.w):
                iother = other[i, j]
                if not iother.isempty():
                    gph_out[i, j] = iother[0, 0]
        return gph_out

    def print(self, frame=False, axes=False, frame_formatter=None):
        grph = deepcopy(self)
        if frame:
            grph = add_frame(grph, frame_formatter)
        if axes:
            grph = add_axes(grph, self.w, self.h)
        print(grph)


def add_axes(grph, w, h):
    # axis order of gratness
    yg = int(math.log10(grph.h) + 1)
    xg = int(math.log10(grph.w) + 1)
    # check offset (caused by frame)
    offset_h = int((grph.h - h) / 2)
    offset_w = int((grph.w - w) / 2)
    # y axis (with frame and axis compensation)
    yaxis = (
        [' ' * yg] * offset_h
        + [str(h - y - 1).zfill(yg) for y in range(h)]
        + [' ' * yg] * offset_h
    )
    # x axis (with frame and axis compensation)
    xaxis = list()
    for i in range(xg):
        n = 10**i
        xaxis.append(
            [' ' * yg]
            + [' '] * offset_w
            + [str(int(x/n) % 10) if (x % n) == 0
               else ' ' for x in range(w)]
            + [' '] * offset_w
        )
    # add to original grid
    grph.grid = np.c_[yaxis, grph.grid]
    grph.grid = np.r_[grph.grid, xaxis]
    return grph


def add_frame(grph, formatter=None):
    chrs = DrawingChars()
    if formatter is not None:
        [chrs.__dict__.update({k: formatter.format(ch)})
         for k, ch in chrs.__dict__.items()]
    li = [[chrs.ur] + grph.w * [chrs.lr] + [chrs.rd]]
    lf = [[chrs.dr] + grph.w * [chrs.lr] + [chrs.ru]]
    grph.grid = np.c_[[chrs.ud] * grph.h, grph.grid, [chrs.ud] * grph.h]
    grph.grid = np.r_[li, grph.grid, lf]
    return grph


def delete_object(obj):
    del obj



from numbers import Number as numeric


class Coordinates:
    def __init__(self, x=0, y=0):
        self.x = int(x)
        self.y = int(y)

    def __add__(self, other):
        return Coordinates(self.x + other[0], self.y + other[1])

    def __sub__(self, other):
        return Coordinates(self.x - other[0], self.y - other[1])

    def __mul__(self, other):
        if isinstance(other, numeric):
            return Coordinates(self.x*other, self.y*other)
        else:
            raise TypeError("'other' must be numeric.")

    def __truediv__(self, other):
        if isinstance(other, numeric):
            return Coordinates(self.x/other, self.y/other)
        else:
            raise TypeError("'other' must be numeric.")

    def __eq__(self, other):
        return (self.x == other.x) & (self.y == other.y)

    def __getitem__(self, item):
        return [self.x, self.y][item]

    def __repr__(self):
        return 'Coordinates({}, {})'.format(self.x, self.y)

    def __floor__(self):
        return Coordinates(int(self.x), int(self.y))

    def __round__(self, n=0):
        return Coordinates(round(self.x, n), round(self.y, n))



# canvas module

from charcad.draw.graph import GraphicObject, GraphicObjectArray, Graph
from charcad.draw.point import Point
from charcad.draw.lines import Route, Horizontal, Vertical
from charcad.draw.text import Text


class Canvas(GraphicObject):
    def __init__(self, w, h):
        super(Canvas, self).__init__()
        self.draw = Draw(self)
        self.w = w
        self.h = h
        self.reset()

    def add_object(self, obj, x=None, y=None, coord=None):
        if not coord is None:
            obj.set_coordinates(coord)
        elif all([x, y]):
            obj.set_coordinates(x, y)
        self.objects.add(obj)

    def remove_object(self, key):
        self.objects.remove(key)

    def reset(self):
        self.objects = GraphicObjectArray()
        self.reset_graph()

    def reset_graph(self):
        self.graph = Graph(w=self.w, h=self.h)

    def show(self, axes=False, frame=False, frame_formatter=None):
        self.update_graph()
        self.graph.print(axes=axes, frame=frame,
                         frame_formatter=frame_formatter)

    def undo(self):
        self.objects.remove(-1)

    def update_graph(self):
        self.reset_graph()
        for obj in self.objects:
            try:
                obj.update()
            except:
                pass
        self.graph.add_objects(self.objects)


class Draw:
    def __init__(self, outer):
        self._canvas = outer
        self.line = self._Line(outer)

    def point(self, p, formatter=None):
        if isinstance(p, (list, tuple)):
            p = Point(*p)
            p.set_formatter(formatter)
        self._canvas.add_object(p)

    def text(self, text, x=0, y=0, transparent=True, formatter=None,
             alignment='center'):
        txt = Text(x, y, text=text, transparent=transparent,
                   formatter=formatter, alignment=alignment)
        self._canvas.add_object(txt)

    class _Line:
        def __init__(self, outer):
            self._canvas = outer

        def __call__(self, direction, *a, **kw):
            if direction.lower() in ['h', 'horizontal']:
                self.horizontal(*a, **kw)
            elif direction.lower() in ['v', 'vertical']:
                self.vertical(*a, **kw)
            else:
                raise Exception(
                    "'%s' is an invalid direction." % direction)

        def horizontal(self, x, y, length, marker=None, formatter=None):
            line = Horizontal(x, y, length, marker=marker, formatter=formatter)
            line.formatter = formatter
            self._canvas.add_object(line)

        def vertical(self, x, y, length, marker=None, formatter=None):
            line = Vertical(x, y, length, marker=marker, formatter=formatter)
            line.formatter = formatter
            self._canvas.add_object(line)

        def route(self, *p, marker='.', origin_marker=None,
                  formatter=None, origin_formatter=None, transparent=True):
            if origin_marker is None:
                origin_marker = marker
            p = list(p)
            for i, item in enumerate(p):
                if isinstance(item, (tuple, list)):
                    p[i] = Point(*item)
                elif isinstance(item, Point):
                    p[i] = item
            route = Route(transparent=transparent)
            route.create_route(*p, marker=marker, origin_marker=origin_marker,
                               formatter=formatter,
                               origin_formatter=origin_formatter)
            self._canvas.add_object(route)



from charcad.draw.coordinates import Coordinates
from charcad.draw.canvas import Canvas
from charcad.draw.graph import add_frame


class Block(Canvas):
    def __init__(self, w, h, x=0, y=0, framed=True, frame_formatter=None):
        super(Block, self).__init__(w=w, h=h)
        self.framed = framed
        self.set_frame_formatter(frame_formatter)
        self.coord = Coordinates(x, y)

    def _add_frame(self):
        self.graph = add_frame(self.graph, formatter=self.frame_formatter)

    def show(self, axes=False):
        self.update()
        self.graph.print(axes=axes, frame=False)

    def set_frame_formatter(self, formatter):
        self.frame_formatter = formatter
        self.update()
    
    def update(self):
        self.reset_graph()
        if self.framed:
            self._add_frame()
        self.graph.add_objects(self.objects)



from charcad.draw.graph import Graph, GraphicObject, GraphicObjectArray
from charcad.draw.point import Point
from charcad.draw.coordinates import Coordinates


class Text(GraphicObject):
    def __init__(self, x=0, y=0, transparent=True, formatter=None,
                 alignment='center', text=None):
        super(Text, self).__init__(x, y, transparent)
        self.formatter = formatter
        self.objects = GraphicObjectArray()
        self.alignment = alignment
        if text is not None:
            self.write(text)

    @property
    def origin(self):
        self.coord = Coordinates(self.x, self.y)
        return self.coord

    def add_point(self, p):
        self.objects.add(p)

    def align(self, alignment=None):
        def calc_offset(w, sz):
            if self.alignment == 'center':
                return int((w - sz)/2)
            elif self.alignment == 'right':
                return w - sz
            elif self.alignment == 'left':
                return 0
        if alignment is None:
            alignment = self.alignment
        else:
            self.alignment = alignment
        w = self.w
        h = self.h
        objects_list = [
            [obj for obj in self.objects if obj.y == i]
            for i in range(h)]
        for lst in objects_list:
            x0 = lst[0].x
            sz = len(lst)
            offset = calc_offset(w, sz) - x0
            for obj in lst:
                obj.coord.x += offset
        self.update_graph()

    def format_text(self, formatter=None):
        if formatter is not None:
            self.formatter = formatter
        for obj in self.objects:
            obj.set_formatter(self.formatter)
        self.update_graph()

    def update_graph(self):
        self.graph = Graph(w=self.w, h=self.h)
        for obj in self.objects:
            self.graph.add_graph(obj)

    def write(self, text):
        if isinstance(text, (list, tuple)):
            w = max([len(t) for t in text])
            h = len(text)
            self.set_y(self.y - (h-1))
            for j, txt in enumerate(text):
                j_ = (h-1) - j
                for i, t in enumerate(txt):
                    p = Point(i, j_, marker=t)
                    self.add_point(p)
        elif isinstance(text, str):
            w = len(text)
            h = 1
            for i, t in enumerate(text):
                p = Point(i, 0, marker=t)
                self.add_point(p)
        else:
            raise TypeError("'text' must be s string or a list of strings")
        self.w = w
        self.h = h
        self.format_text()
        self.align()



import math
from numbers import Number as numeric

from charcad.draw.coordinates import Coordinates
from charcad.draw.utils import calc_angle


class Vector:
    def __init__(self, *p, origin=Coordinates(0, 0)):
        if isinstance(p[0], tuple):
            p = Coordinates(*p[0])
        elif isinstance(p[0], Coordinates):
            p = p[0]
        elif isinstance(p[0], numeric):
            p = Coordinates(*p)

        if isinstance(origin, tuple):
            origin = Coordinates(*origin)

        self.p = p
        self.origin = origin

    def __abs__(self):
        return math.hypot(*self.p)

    def __add__(self, other):
        return Vector(self.p + other.p, self.origin)

    def __mul__(self, other):
        if isinstance(other, (float, int)):
            return Vector(self.p * other)
        else:
            raise Exception("'other' must be scalar.")

    def __truediv__(self, other):
        if isinstance(other, (float, int)):
            return Vector(self.p / other)
        else:
            raise Exception("'other' must be scalar.")

    def __repr__(self):
        return "vector[{}, {}]".format(self.p, self.origin)

    def angle(self, other=None):
        if other is None:
            other = Vector(1, 0)
        return calc_angle(self, other)

    def dot(self, other):
        return sum([i1 * i2 for i1, i2 in zip(self.p, other.p)])

    @property
    def unit(self):
        return Vector(self.p/abs(self))



import math


# properties

class DrawingChars:
    def __init__(self):
        self.ud = '│'
        self.ur = '┌'
        self.dr = '└'
        self.ru = '┘'
        self.rd = '┐'
        self.lr = '─'

        self.ilr = '_'
        self.oin = '┼'
        self.din = '╳'
        self.dru = '╱'
        self.drd = '╲'


chrs = DrawingChars()


# fuctions

def calc_angle(v1, v2):
    return math.acos(clamp(v1.dot(v2)/abs(v1)*abs(v2), 0, 1))


def calc_distance(p1, p2, factors=(1, 1)):
    dx = (p2[0] - p1[0]) * factors[0]
    dy = (p2[1] - p1[1]) * factors[1]
    return math.hypot(dx, dy)


def clamp(val, minval, maxval):
    return max(min(val, maxval), minval)


def force_list(var):
    if isinstance(var, (int, float)):
        var = [var]
    elif not isinstance(var, list):
        var = list(var)
    return var


# text formatters

def assign_aspect(value, aspect):
    def deco(formatter):
        def wrapper():
            if aspect == 'style':
                formatter.style = value
            if aspect == 'color':
                formatter.color = value
            if aspect == 'bgcolor':
                if formatter.color == '38':
                    formatter.color = '30'
                formatter.background_color = value
        return wrapper
    return deco


ESCAPE = ['\033[', '\033[0m']
STYLES = {
    'none':         assign_aspect('0', 'style'),
    'bold':         assign_aspect('1', 'style'),
    'negative':     assign_aspect('2', 'style'),
    'negative2':    assign_aspect('5', 'style'),
    'underline':    assign_aspect('4', 'style'),
}
COLORS = {
    'black':    assign_aspect('30', 'color'),
    'red':      assign_aspect('31', 'color'),
    'green':    assign_aspect('32', 'color'),
    'yellow':   assign_aspect('33', 'color'),
    'blue':     assign_aspect('34', 'color'),
    'purple':   assign_aspect('35', 'color'),
    'cyan':     assign_aspect('36', 'color'),
    'white':    assign_aspect('37', 'color'),
    'none':     assign_aspect('38', 'color'),
}
BACKGROUND_COLORS = {
    'black':    assign_aspect('40m', 'bgcolor'),
    'red':      assign_aspect('41m', 'bgcolor'),
    'green':    assign_aspect('42m', 'bgcolor'),
    'yellow':   assign_aspect('43m', 'bgcolor'),
    'blue':     assign_aspect('44m', 'bgcolor'),
    'purple':   assign_aspect('45m', 'bgcolor'),
    'cyan':     assign_aspect('46m', 'bgcolor'),
    'white':    assign_aspect('47m', 'bgcolor'),
    'none':     assign_aspect('48m', 'bgcolor'),
}


class Formatter:
    class _AspectSetter:
        def __init__(self, outer, aspect_dict):
            aspects = aspect_dict.copy()
            [aspects.update({key: val(outer)}) for key, val in aspects.items()]
            self.__dict__ = aspects

    def __init__(self):
        # init aspacts
        self.style = None
        self.color = None
        self.background_color = None
        # formats
        self.set_style = self._AspectSetter(self, STYLES)
        self.set_color = self._AspectSetter(self, COLORS)
        self.set_background_color = self._AspectSetter(self, BACKGROUND_COLORS)
        # config
        self.set_style.none()
        self.set_color.none()
        self.set_background_color.none()

    def __repr__(self):
        return self.format('Hello! I am Formatter ' + str(id(self)))

    def format(self, text):
        fmt = [self.style, self.color, self.background_color]
        if all(fmt):
            text_out = (';'.join(fmt) + text).join(ESCAPE)
        else:
            text_out = text
        return text_out


# decorators



from setuptools import setup, find_packages

exec(open('charcad/version.py', 'r').read())

VERSION = __version__

requirements = open('requirements.txt', 'r').read().splitlines()

setup(
    name='charcad',
    packages=find_packages(),
    version=VERSION,
    license='MIT',
    description='Character-based drawing tool',
    author='Victor Zoni',
    author_email='vczoni@gmail.com',
    url='https://github.com/vczoni/charcad',
    download_url=VERSION.join(
        ['https://github.com/vczoni/charcad/archive/v', '.tar.gz']),
    keywords=['CHARACTER', 'DRAWING', 'BASIC'],
    install_requires=requirements,
    python_requires='>=3',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)




from charcad.draw.coordinates import Coordinates
from charcad.draw.graph import GraphicObject, GraphicObjectArray, Graph
from charcad.draw.point import Point
from charcad.draw.utils import calc_angle, calc_distance, chrs as ch


class Lines(GraphicObject):
    def __init__(self, x=0, y=0, transparent=True, formatter=None):
        super(Lines, self).__init__(x, y, transparent)
        self.formatter = formatter
        self.objects = GraphicObjectArray()


class Line(Lines):
    def __init__(self, x, y, formatter=None):
        super(Line, self).__init__(x, y, formatter=formatter)

    def __repr__(self):
        return 'Line at <%s>' % id(self)

    def __len__(self):
        return len(self.objects)

    @property
    def origin(self):
        self.coord = Coordinates(self.x, self.y)
        return self.coord

    @property
    def end(self):
        x = self.objects[-1].coord.x
        y = self.objects[-1].coord.y
        self.coord = Coordinates(x, y)
        return self.coord

    @property
    def length(self):
        return len(self)

    def add_point(self, p):
        self.objects.add(p)

    def update_graph(self, w=1, h=1):
        self.graph = Graph(w=w, h=h)
        origin = self.origin
        for obj in self.objects:
            obj.coord -= origin
            self.graph.add_graph(obj)


class Horizontal(Line):
    def __init__(self, x, y, length, marker=None, formatter=None):
        super(Horizontal, self).__init__(x, y, formatter)
        if marker is None:
            marker = ch.lr
        self.marker = marker
        self.create(length)

    def create(self, length):
        for i in range(length):
            p = Point(self.x+i, self.y, marker=self.marker,
                      formatter=self.formatter)
            self.add_point(p)
        self.update_graph(w=length)


class Vertical(Line):
    def __init__(self, x, y, length, marker=None, formatter=None):
        super(Vertical, self).__init__(x, y, formatter)
        if marker is None:
            marker = ch.ud
        self.marker = marker
        self.create(length)

    def create(self, length):
        for i in range(length):
            p = Point(self.x, self.y+i, marker=self.marker,
                      formatter=self.formatter)
            self.add_point(p)
        self.update_graph(h=length)


class Route(Lines):
    def __init__(self, transparent=True):
        super(Route, self).__init__(transparent=transparent)
        self.objects = GraphicObjectArray()

    def __repr__(self):
        return 'Route at <%s>' % id(self)

    def __len__(self):
        return len(self.objects)

    @property
    def origin(self):
        x = min([obj.coord.x for obj in self.objects])
        y = min([obj.coord.y for obj in self.objects])
        return Coordinates(x, y)

    @property
    def h(self):
        return (max([obj.coord.y for obj in self.objects])
                - min([obj.coord.y for obj in self.objects])
                + 1)

    @property
    def w(self):
        return (max([obj.coord.x for obj in self.objects])
                - min([obj.coord.x for obj in self.objects])
                + 1)

    def add_point(self, p):
        self.objects.add(p)

    def create_route(self, *points, marker='.', origin_marker='.',
                     formatter=None, origin_formatter=None):
        for point, next_point in zip(points[0:-1], points[1:]):
            point.set_marker(marker)
            point.set_formatter(formatter)
            self.connect(point, next_point)
        # edit origins
        self.objects[0].set_marker(origin_marker)
        self.objects[0].set_formatter(origin_formatter)
        self.objects[-1].set_marker(origin_marker)
        self.objects[-1].set_formatter(origin_formatter)
        self.coord = self.origin
        self.update()

    def connect(self, p1, p2):
        current_point = p1
        target_point = p2
        arrived = current_point == target_point
        self.add_point(current_point)
        while not arrived:
            next_points = [current_point + m for m in movements]
            distances = [calc_distance(p, target_point)
                         for p in next_points]
            idx = distances.index(min(distances))
            current_point += movements[idx]
            self.add_point(current_point)
            arrived = current_point == target_point

    def update(self):
        self.graph = Graph(w=self.w, h=self.h)
        origin = self.origin
        for obj in self.objects:
            obj.coord -= origin
            self.graph.add_graph(obj)


movements = [
    Coordinates(0, 1),
    Coordinates(1, 1),
    Coordinates(1, 0),
    Coordinates(1, -1),
    Coordinates(0, -1),
    Coordinates(-1, -1),
    Coordinates(-1, 0),
    Coordinates(-1, 1),
]
